#ifndef _LIDAR_LITE_H
#define _LIDAR_LITE_H

#include <iostream>
#include <sstream>  
#include <string>    
#include <iomanip>
#include <cmath>
#include <usb_i2c.h>
#include "exceptions/lidar_lite_exceptions.h"

// The device has a 7-bit slave address with a default value of 0x62.
// Fill in new address here if changed. 
// See operating manual for instructions.
unsigned char DEVICE_ID = 0x62;

enum V3_REGISTER_DEF
{  
	ACQ_COMMAND = 			0x00,	// Device command.
	STATUS = 						0x01,	// System status.
	SIG_COUNT_VAL = 		0x02,	// Maximum acquisition count.
	ACQ_CONFIG_REG = 		0x04,	// Acquisition mode control.
	VELOCITY = 					0x09,	// Velocity measurement output.
	PEAK_CORR = 				0x0c,	// Peak value in correlation record.
	NOISE_PEAK = 				0x0d,	// Correlation record noise floor.
	SIGNAL_STRENGTH = 	0x0e,	// Received signal strength.
	FULL_DELAY_HIGH = 	0x0f,	// Distance measurement high byte.
	FULL_DELAY_LOW = 		0x10,	// Distance measurement low byte.
	OUTER_LOOP_COUNT = 	0x11,	// Burst measurement count control.
	REF_COUNT_VAL = 		0x12,	// Reference acquisition count.
	LAST_DELAY_HIGH = 	0x14,	// Previous distance measurement high byte.
	LAST_DELAY_LOW = 		0x15,	// Previous distance measurement low byte.
	UNIT_ID_HIGH = 			0x16,	// Serial number high byte.
	UNIT_ID_LOW = 			0x17,	// Serial number low byte.
	I2C_ID_HIGH = 			0x18,	// Write serial number high byte for I2C address unlock.
	I2C_ID_LOW = 				0x19,	// Write serial number low byte for I2C address unlock.
	I2C_SEC_ADDR = 			0x1a,	// Write new I2C address after unlock.
	THRESHOLD_BYPASS = 	0x1c,	// Peak detection threshold bypass.
	I2C_CONFIG = 				0x1e,	// Default address response control.
	COMMAND = 					0x40,	// State command.
	MEASURE_DELAY = 		0x45,	// Delay between automatic measurements.
	PEAK_BCK = 					0x4c,	// Second largest peak value in correlation record.
	CORR_DATA = 				0x52,	// Correlation record data low byte.
	CORR_DATA_SIGN = 		0x53,	// Correlation record data high byte.
	ACQ_SETTINGS =			0x5d,	// Correlation record memory bank select.
	POWER_CONTROL =			0x65,	// Power state control.
};

enum DEV_STATUS 
{  
	IDLE,
	RUNNING
};

class CLidarLite
{
  private:

		CUSBI2C* adapter_;			// Device object.
		std::string serial_;		// Serial port (e.g., run dmesg).
		int config_;						// Device configuration option (see Class header).
		int status_;						// Device status.
		int max_unbiases_meas_;	// Maximum of unbiased readings (not more than 100).

		// Blinks the usb_i2c adapter led.	
		void blink_led_usb_adapter(void);

		// Converts an HEX number (int object) to string. Print purposes.
		std::string hex_to_str(const int &val);

		// Writes a command (cmd) to the register of the I2C device (addr).
		void write(unsigned char addr, unsigned char cmd);

		// Read byte/s (len=[1,2]) from the register of the I2C device (addr).
		int read(unsigned char addr, int len);

    // Successive 2-byte readings (Autoincrement of address: A note about 0x8f vs 0x0f)
    // Set the highest bit of any register to "1" if you set the high byte of a register 
    // and then take succesive readings from that register, then LIDAR-Lite automatically 
    // increments the register one for each read. 
    // An example: If we want to read the high and low bytes for the distance, we could 
    // take two single readings from 0x0f and 0x10, or we could take 2 byte read from 
    // register 0x8f. 0x8f = 10001111 and 0x0f = 00001111, meaning that 0x8f is 0x0f with 
    // the high byte set to "1", ergo it autoincrements.
		unsigned short int addr_to_read_2bytes(unsigned char first_byte_addr);

	  // Configure Lidar Lite device with parameters from Class constructor
	  void config_dev(void);

  public:

  	//	- Inputs:
  	//		- serial: Device serial ID (e.g. run 'dmesg').
  	// 		- config:	Lidar Lite configuration:  Default 0.
  	//   			0: Default mode, balanced performance.
  	//	   		1: Short range, high speed. Uses 0x1d maximum acquisition count.
  	//   			2: Default range, higher speed short range. Turns on quick termination
  	//       	 	 detection for faster measurements at short range (with decreased
  	//           accuracy)
  	//   			3: Maximum range. Uses 0xff maximum acquisition count.
  	//   			4: High sensitivity detection. Overrides default valid measurement detection
  	//       	   algorithm, and uses a threshold value for high sensitivity and noise.
  	//   			5: Low sensitivity detection. Overrides default valid measurement detection
  	//           algorithm, and uses a threshold value for low sensitivity and noise.	
    CLidarLite(const std::string &serial, const int config = 0);
    ~CLidarLite(void);

		void open(void);
		void close(void);
		void reset(void);

		// Set new device configuration (see constructor for details on modes)
		void set_new_config(const int config);

	  // Get range
	  // Take a distance measurement and return the result.
		//	- Inputs:
	  // 		- biasCorrection: Default true. Take aquisition with receiver bias
	  // 			correction. If set to false measurements will be faster. Receiver bias
	  // 			correction must be performed periodically. (e.g. 1 out of every 100
	  // 			readings).
		int get_range(bool biasCorrection = true);
};

#endif
